// Copyright 2006 Benedikt Böhm <hollow@gentoo.org>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the
// Free Software Foundation, Inc.,
// 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/*!
 * @defgroup list Simple doubly linked lists
 *
 * The simplest kind of linked list is a singly-linked list, which has one link
 * per node. This link points to the next node in the list, or to a null value
 * or empty list if it is the final node; e.g. 12 -> 99 -> 37 -> NULL.
 *
 * A more sophisticated kind of linked list is a doubly-linked list. Each node
 * has two links: one points to the previous node, or points to a null value or
 * empty list if it is the first node; and one points to the next, or points to
 * a null value or empty list if it is the final node; e.g.
 * NULL <- 26 <-> 56 <-> 46 -> NULL.
 *
 * The list family of functions and macros provide routines to create a list,
 * add, move or remove elements and iterate over the list.
 *
 * @{
 */

#ifndef _LUCID_LIST_H
#define _LUCID_LIST_H

#include <stddef.h>

/*! @brief get container of list head */
#define container_of(ptr, type, member) \
	((type *)((char *)(ptr) - offsetof(type, member)))

/*
 * Simple doubly linked list implementation.
 *
 * Some of the internal functions ("__xxx") are useful when
 * manipulating whole lists rather than single entries, as
 * sometimes we already know the next/prev entries and we can
 * generate better code by using them directly rather than
 * using the generic single-entry routines.
 */

/*! @brief list head */
struct list_head {
	struct list_head *next, *prev;
};

/*! @brief list head initialization */
#define LIST_HEAD_INIT(name) { &(name), &(name) }

/*! @brief list head creation */
#define LIST_HEAD(name) \
	struct list_head name = LIST_HEAD_INIT(name)

static inline void INIT_LIST_HEAD(struct list_head *list)
{
	list->next = list;
	list->prev = list;
}

/*
 * Insert a new entry between two known consecutive entries.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static inline
void __list_add(struct list_head *new,
                struct list_head *prev,
                struct list_head *next)
{
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}

/*!
 * @brief add a new entry
 *
 * Insert a new entry after the specified head.
 * This is good for implementing stacks.
 *
 * @param new  new entry to be added
 * @param head list head to add it after
 */
static inline
void list_add(struct list_head *new, struct list_head *head)
{
	__list_add(new, head, head->next);
}

/*!
 * @brief add a new entry
 *
 * Insert a new entry before the specified head.
 * This is useful for implementing queues.
 *
 * @param new  new entry to be added
 * @param head list head to add it before
 */
static inline
void list_add_tail(struct list_head *new, struct list_head *head)
{
	__list_add(new, head->prev, head);
}

/*
 * Delete a list entry by making the prev/next entries
 * point to each other.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static inline
void __list_del(struct list_head * prev, struct list_head * next)
{
	next->prev = prev;
	prev->next = next;
}

/*!
 * @brief deletes entry from list
 *
 * @param entry the element to delete from the list
 *
 * @note list_empty on entry does not return true after this, the entry is
 *       in an undefined state
 */
static inline
void list_del(struct list_head *entry)
{
	__list_del(entry->prev, entry->next);
	entry->next = (void *) 0;
	entry->prev = (void *) 0;
}

/*!
 * @brief deletes entry from list and reinitialize it
 *
 * @param entry the element to delete from the list
 */
static inline
void list_del_init(struct list_head *entry)
{
	__list_del(entry->prev, entry->next);
	INIT_LIST_HEAD(entry);
}

/*!
 * @brief delete from one list and add as another's head
 *
 * @param list the entry to move
 * @param head the head that will precede our entry
 */
static inline
void list_move(struct list_head *list, struct list_head *head)
{
	__list_del(list->prev, list->next);
	list_add(list, head);
}

/*!
 * @brief delete from one list and add as another's tail
 *
 * @param list the entry to move
 * @param head the head that will follow our entry
 */
static inline
void list_move_tail(struct list_head *list, struct list_head *head)
{
	__list_del(list->prev, list->next);
	list_add_tail(list, head);
}

/*!
 * @brief tests whether a list is empty
 *
 * @param head the list to test
 */
static inline
int list_empty(const struct list_head *head)
{
	return head->next == head;
}

static inline
void __list_splice(struct list_head *list, struct list_head *head)
{
	struct list_head *first = list->next;
	struct list_head *last = list->prev;
	struct list_head *at = head->next;
	
	first->prev = head;
	head->next = first;
	
	last->next = at;
	at->prev = last;
}

/*!
 * @brief join two lists
 *
 * @param list the new list to add
 * @param head the place to add it in the first list
 */
static inline
void list_splice(struct list_head *list, struct list_head *head)
{
	if (!list_empty(list))
		__list_splice(list, head);
}

/*!
 * @brief join two lists and reinitialise the emptied list
 *
 * @param list the new list to add
 * @param head the place to add it in the first list
 */
static inline
void list_splice_init(struct list_head *list, struct list_head *head)
{
	if (!list_empty(list)) {
		__list_splice(list, head);
		INIT_LIST_HEAD(list);
	}
}

/*!
 * @brief get the struct for this entry
 *
 * @param ptr    the &struct list_head pointer
 * @param type   the type of the struct this is embedded in
 * @param member the name of the list_struct within the struct
 */
#define list_entry(ptr, type, member) \
	container_of(ptr, type, member)

/*!
 * @brief iterate over a list
 *
 * @param pos  the &struct list_head to use as a loop counter
 * @param head the head for your list
 */
#define list_for_each(pos, head) \
	for (pos = (head)->next; pos != (head); pos = pos->next)

/*!
 * @brief iterate over a list backwards
 *
 * @param pos  the &struct list_head to use as a loop counter
 * @param head the head for your list
 */
#define list_for_each_prev(pos, head) \
	for (pos = (head)->prev; pos != (head); pos = pos->prev)

/*!
 * @brief iterate over a list safe against removal of list entry
 *
 * @param pos   the &struct list_head to use as a loop counter
 * @param n     another &struct list_head to use as temporary storage
 * @param head  the head for your list
 */
#define list_for_each_safe(pos, n, head) \
	for (pos = (head)->next, n = pos->next; pos != (head); \
		pos = n, n = pos->next)

/*!
 * @brief iterate over list of given type
 *
 * @param pos    the type * to use as a loop counter
 * @param head:  the head for your list
 * @param member the name of the list_struct within the struct
 */
#define list_for_each_entry(pos, head, member) \
	for (pos = list_entry((head)->next, typeof(*pos), member); \
	     &pos->member != (head); \
	     pos = list_entry(pos->member.next, typeof(*pos), member))

/*!
 * @brief iterate backwards over list of given type.
 *
 * @param pos    the type * to use as a loop counter
 * @param head   the head for your list
 * @param member the name of the list_struct within the struct
 */
#define list_for_each_entry_reverse(pos, head, member) \
	for (pos = list_entry((head)->prev, typeof(*pos), member); \
	     &pos->member != (head); \
	     pos = list_entry(pos->member.prev, typeof(*pos), member))

/*!
 * @brief iterate over list of given type safe against removal of list entry
 *
 * @param pos    the type * to use as a loop counter
 * @param n      another type * to use as temporary storage
 * @param head   the head for your list
 * @param member the name of the list_struct within the struct
 */
#define list_for_each_entry_safe(pos, n, head, member) \
	for (pos = list_entry((head)->next, typeof(*pos), member), \
	     n = list_entry(pos->member.next, typeof(*pos), member); \
	     &pos->member != (head); \
	     pos = n, n = list_entry(n->member.next, typeof(*n), member))

/*!
 * @brief iterate backwards over list of given type safe against removal of list entry
 *
 * @param pos    the type * to use as a loop counter
 * @param n      another type * to use as temporary storage
 * @param head   the head for your list
 * @param member the name of the list_struct within the struct
 */
#define list_for_each_entry_safe_reverse(pos, n, head, member) \
	for (pos = list_entry((head)->prev, typeof(*pos), member), \
	     n = list_entry(pos->member.prev, typeof(*pos), member); \
	     &pos->member != (head); \
	     pos = n, n = list_entry(n->member.prev, typeof(*n), member))

#endif

/*! @} list */
